#!/usr/bin/env node


/**
 * Copyright 2015-present Greg Hurrell. All rights reserved.
 * Licensed under the terms of the MIT license.
 *
 * 
 */

'use strict';

var _slicedToArray = (function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i['return']) _i['return'](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError('Invalid attempt to destructure non-iterable instance'); } }; })();

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i]; return arr2; } else { return Array.from(arr); } }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

require('babel/polyfill');

var _bluebird = require('bluebird');

var _bluebird2 = _interopRequireDefault(_bluebird);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _yargs = require('yargs');

var _yargs2 = _interopRequireDefault(_yargs);

var readFile = _bluebird2['default'].promisify(_fs2['default'].readFile);

/**
 * Allows us to generate intermediate layouts with unique names.
 */
var layoutCounter = 0;

process.on('unhandledRejection', function (reason) {
  throw reason;
});

var log = console.log.bind(console);

function print(string) {
  process.stdout.write(string);
}

/**
 * The center of each key is measured in an arbitrary coordinate system (pixels)
 * based on this image:
 *
 * https://support.apple.com/library/content/dam/edam/applecare/images/en_US/keyboards/english_notebook.png
 */
var KEYS = [
// Row 0.
{ id: 0, row: 0, column: 0, name: '⎋', x: 126, y: 71 }, { id: 1, row: 0, column: 1, name: 'F1', x: 352, y: 71 }, { id: 2, row: 0, column: 2, name: 'F2', x: 590, y: 71 }, { id: 3, row: 0, column: 3, name: 'F3', x: 816, y: 71 }, { id: 4, row: 0, column: 4, name: 'F4', x: 1052, y: 71 }, { id: 5, row: 0, column: 5, name: 'F5', x: 1286, y: 71 }, { id: 6, row: 0, column: 6, name: 'F6', x: 1514, y: 71 }, { id: 7, row: 0, column: 7, name: 'F7', x: 1750, y: 71 }, { id: 8, row: 0, column: 8, name: 'F8', x: 1978, y: 71 }, { id: 9, row: 0, column: 9, name: 'F9', x: 2216, y: 71 }, { id: 10, row: 0, column: 10, name: 'F10', x: 2450, y: 71 }, { id: 11, row: 0, column: 11, name: 'F11', x: 2678, y: 71 }, { id: 12, row: 0, column: 12, name: 'F12', x: 2914, y: 71 }, { id: 13, row: 0, column: 13, name: '⌽', x: 3136, y: 71 },

// Row 1.
{ id: 14, row: 1, column: 0, name: '~', x: 115, y: 254 }, { id: 15, row: 1, column: 1, name: '1', x: 339, y: 254 }, { id: 16, row: 1, column: 2, name: '2', x: 573, y: 254 }, { id: 17, row: 1, column: 3, name: '3', x: 789, y: 254 }, { id: 18, row: 1, column: 4, name: '4', x: 1015, y: 254 }, { id: 19, row: 1, column: 5, name: '5', x: 1239, y: 254 }, { id: 20, row: 1, column: 6, name: '6', x: 1467, y: 254 }, { id: 21, row: 1, column: 7, name: '7', x: 1691, y: 254 }, { id: 22, row: 1, column: 8, name: '8', x: 1915, y: 254 }, { id: 23, row: 1, column: 9, name: '9', x: 2141, y: 254 }, { id: 24, row: 1, column: 10, name: '0', x: 2359, y: 254 }, { id: 25, row: 1, column: 11, name: '-', x: 2585, y: 254 }, { id: 26, row: 1, column: 12, name: '=', x: 2813, y: 254 }, { id: 27, row: 1, column: 13, name: '⌫', x: 3092, y: 254 },

// Row 2.
{ id: 28, row: 2, column: 0, name: '⇥', x: 177, y: 474 }, { id: 29, row: 2, column: 1, name: 'q', x: 455, y: 474 }, { id: 30, row: 2, column: 2, name: 'w', x: 679, y: 474 }, { id: 31, row: 2, column: 3, name: 'e', x: 901, y: 474 }, { id: 32, row: 2, column: 4, name: 'r', x: 1123, y: 474 }, { id: 33, row: 2, column: 5, name: 't', x: 1351, y: 474 }, { id: 34, row: 2, column: 6, name: 'y', x: 1573, y: 474 }, { id: 35, row: 2, column: 7, name: 'u', x: 1799, y: 474 }, { id: 36, row: 2, column: 8, name: 'i', x: 2025, y: 474 }, { id: 37, row: 2, column: 9, name: 'o', x: 2245, y: 474 }, { id: 38, row: 2, column: 10, name: 'p', x: 2471, y: 474 }, { id: 39, row: 2, column: 11, name: '[', x: 2695, y: 474 }, { id: 40, row: 2, column: 12, name: ']', x: 2921, y: 474 }, { id: 41, row: 2, column: 13, name: '\\', x: 3148, y: 474 },

// Row 3 (home row).
{ id: 42, row: 3, column: 0, name: '⇪', x: 196, y: 692 }, { id: 43, row: 3, column: 1, name: 'a', x: 510, y: 692 }, { id: 44, row: 3, column: 2, name: 's', x: 736, y: 692 }, { id: 45, row: 3, column: 3, name: 'd', x: 958, y: 692 }, { id: 46, row: 3, column: 4, name: 'f', x: 1180, y: 692 }, { id: 47, row: 3, column: 5, name: 'g', x: 1406, y: 692 }, { id: 48, row: 3, column: 6, name: 'h', x: 1631, y: 692 }, { id: 49, row: 3, column: 7, name: 'j', x: 1857, y: 692 }, { id: 50, row: 3, column: 8, name: 'k', x: 2079, y: 692 }, { id: 51, row: 3, column: 9, name: 'l', x: 2301, y: 692 }, { id: 52, row: 3, column: 10, name: ';', x: 2531, y: 692 }, { id: 53, row: 3, column: 11, name: '\'', x: 2749, y: 692 }, { id: 54, row: 3, column: 12, name: '↩', x: 3057, y: 692 },

// Row 4.
{ id: 55, row: 4, column: 0, name: '⇧ (Left)', x: 259, y: 910 }, { id: 56, row: 4, column: 1, name: 'z', x: 619, y: 910 }, { id: 57, row: 4, column: 2, name: 'x', x: 843, y: 910 }, { id: 58, row: 4, column: 3, name: 'c', x: 1071, y: 910 }, { id: 59, row: 4, column: 4, name: 'v', x: 1293, y: 910 }, { id: 60, row: 4, column: 5, name: 'b', x: 1519, y: 910 }, { id: 61, row: 4, column: 6, name: 'n', x: 1743, y: 910 }, { id: 62, row: 4, column: 7, name: 'm', x: 1965, y: 910 }, { id: 63, row: 4, column: 8, name: ',', x: 2191, y: 910 }, { id: 64, row: 4, column: 9, name: '.', x: 2421, y: 910 }, { id: 65, row: 4, column: 10, name: '/', x: 2643, y: 910 }, { id: 66, row: 4, column: 11, name: '⇧ (Right)', x: 3007, y: 910 },

// Row 5 (never analyzed, but here for completeness).
{ id: 67, row: 5, column: 0, name: 'fn', x: 115, y: 1138 }, { id: 68, row: 5, column: 1, name: '⌃ (Left)', x: 333, y: 1138 }, { id: 68, row: 5, column: 1, name: '⌥ (Left)', x: 565, y: 1138 }, { id: 68, row: 5, column: 1, name: '⌘ (Left)', x: 817, y: 1138 }, { id: 68, row: 5, column: 1, name: '␣', x: 1517, y: 1138 }, { id: 68, row: 5, column: 1, name: '⌘ (Right)', x: 2221, y: 1138 }, { id: 68, row: 5, column: 1, name: '⌥ (Right)', x: 2469, y: 1138 }, { id: 68, row: 5, column: 1, name: '←', x: 2697, y: 1200 }, // half-height
{ id: 68, row: 5, column: 1, name: '↑', x: 2919, y: 1082 }, // half-height
{ id: 68, row: 5, column: 1, name: '↓', x: 2919, y: 1200 }, // half-height
{ id: 68, row: 5, column: 1, name: '→', x: 3148, y: 1200 }];

/**
 * Keys marked with a 1 won't be considered for moves.
 *
 * For now, only consider moving letters, except for "A", which I want to stay
 * right where it is because its my tmux prefix.
 */
// half-height
var MASK = [
/* Row 0: */1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
/* Row 1: */1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
/* Row 2: */1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
/* Row 3: */1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
/* Row 4: */1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
/* Row 5: */1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1];

var ROWS = [
/* Row 0: */'F-keys',
/* Row 1: */'Number',
/* Row 2: */'Top',
/* Row 3: */'Middle',
/* Row 4: */'Bottom',
/* Row 5: */'Modifers/Space'];

var FINGER_NAMES = [
/* 0 */'Left Pinkie',
/* 1 */'Left Ring Finger',
/* 2 */'Left Middle Finger',
/* 3 */'Left Index Finger',
/* 4 */'Left Thumb',
/* 5 */'Right Thumb',
/* 6 */'Right Index Finger',
/* 7 */'Right Middle Finger',
/* 8 */'Right Ring Finger',
/* 9 */'Right Pinkie'];

/**
 * My shockingly unorthodox finger-to-key mappings.
 *
 * Note the almost non-existent use of pinkies, and relatively heavy use of the
 * ring fingers, which are typically regarded as among the least flexible
 * fingers.
 */
var FINGERS_PLACEMENTS = [
/* Row 0: */1, 2, 2, 3, 3, 3, 6, 6, 6, 7, 7, 8, 8, 8,
/* Row 1: */1, 1, 2, 3, 3, 3, 6, 6, 6, 7, 7, 7, 8, 8,
/* Row 2: */1, 1, 1, 2, 3, 3, 6, 6, 7, 7, 8, 8, 8, 8,
/* Row 3: */1, 1, 1, 2, 3, 3, 6, 6, 7, 8, 8, 8, 8,
/* Row 4: */0, 1, 2, 3, 3, 6, 6, 6, 7, 8, 8, 9,
/* Row 5: */0, 0, 4, 4, 5, 5, 5, 6, 7, 7, 8];

/**
 * The key on which each finger is considered to be in its optimal resting
 * position (ie on the home row).
 */
var FINGER_HOME_KEYS = [
/* 0: left pinkie */3 * 14 + 1,
/* 1: left ring */3 * 14 + 2,
/* 2: left middle */3 * 14 + 3,
/* 3: left index */3 * 14 + 4,
/* 4: left thumb */3 * 14 + 1 * 13 + 5,
/* 5: right thumb */3 * 14 + 1 * 13 + 5,
/* 6: right index */3 * 14 + 7,
/* 7: right middle */3 * 14 + 8,
/* 8: right ring */3 * 14 + 9,
/* 9: right pinkie */3 * 14 + 10];

var FINGER_STRENGTHS = [
/* 0: left pinkie */0.1,
/* 1: left ring */0.6,
/* 2: left middle */0.7,
/* 3: left index */0.9,
/* 4: left thumb */0.4,
/* 5: right thumb */0.5,
/* 6: right index */1.0,
/* 7: right middle */0.8,
/* 8: right ring */0.7,
/* 9: right pinkie */0.1];

/**
 * This is the "human-readable" version of the layout.
 *
 * Array items are used to represent shifted keys (the first element being the
 * unshifted version and the second element being the shifted version).
 */
var LAYOUTS = {
  COLEMAK: {
    keys: [
    /* Row 0: */'⎋', 'F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10', 'F11', 'F12', '⌽',
    /* Row 1: */['`', '~'], ['1', '!'], ['2', '@'], ['3', '#'], ['4', '$'], ['5', '%'], ['6', '^'], ['7', '&'], ['8', '*'], ['9', '('], ['0', ')'], ['-', '_'], ['=', '+'], '⌫',
    /* Row 2: */'⇥', 'q', 'w', 'f', 'p', 'g', 'j', 'l', 'u', 'y', [';', ':'], ['[', '{'], [']', '}'], ['\\', '|'],
    /* Row 3: */'⇪', 'a', 'r', 's', 't', 'd', 'h', 'n', 'e', 'i', 'o', ["'", '"'], '↩',
    /* Row 4: */'⇧ (Left)', 'z', 'x', 'c', 'v', 'b', 'k', 'm', [',', '<'], ['.', '>'], ['/', '?'], '⇧ (Right)',
    /* Row 5: */'fn', '⌃ (Left)', '⌥ (Left)', '⌘ (Left)', '␣', '⌘ (Right)', '⌥ (Right)', '←', '↑', '↓', '→'],
    name: 'Colemak'
  },

  QWERTY: {
    keys: [
    /* Row 0: */'⎋', 'F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10', 'F11', 'F12', '⌽',
    /* Row 1: */['`', '~'], ['1', '!'], ['2', '@'], ['3', '#'], ['4', '$'], ['5', '%'], ['6', '^'], ['7', '&'], ['8', '*'], ['9', '('], ['0', ')'], ['-', '_'], ['=', '+'], '⌫',
    /* Row 2: */'⇥', 'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', ['[', '{'], [']', '}'], ['\\', '|'],
    /* Row 3: */'⇪', 'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', [';', ':'], ["'", '"'], '↩',
    /* Row 4: */'⇧ (Left)', 'z', 'x', 'c', 'v', 'b', 'n', 'm', [',', '<'], ['.', '>'], ['/', '?'], '⇧ (Right)',
    /* Row 5: */'fn', '⌃ (Left)', '⌥ (Left)', '⌘ (Left)', '␣', '⌘ (Right)', '⌥ (Right)', '←', '↑', '↓', '→'],
    name: 'Qwerty'
  }
};

var layoutLookupMaps = new Map();

/**
 * Given a "human-readable" layout, return a map for fast look-up from input
 * character to pressed key.
 */
function getLayoutLookupMap(layout) {
  if (!layoutLookupMaps.has(layout.name)) {
    (function () {
      var map = {};
      layout.keys.forEach(function (key, index) {
        if (Array.isArray(key)) {
          map[key[0]] = { index: index, shift: false };
          map[key[1]] = { index: index, shift: true };
        } else {
          map[key] = { index: index, shift: false };
        }
      });
      layoutLookupMaps.set(layout.name, map);
    })();
  }
  return layoutLookupMaps.get(layout.name);
}

/**
 * Returns the "distance" (in arbitrary units) between keys `a` and `b`.
 */
function getDistance(a, b) {
  // Via Pythagorean theorem (a^2 + b^2 = c^2).
  return Math.sqrt(Math.pow(Math.abs(a.x - b.x), 2) + Math.pow(Math.abs(a.y - b.y), 2));
}

/**
 * To compute normalized distances, we need to know the maximum distance of a
 * finger from its home key.
 */
var maximumDistanceFromHomeKey = Math.max.apply(Math, _toConsumableArray(FINGERS_PLACEMENTS.map(function (placement, i) {
  return getDistance(KEYS[i], KEYS[FINGER_HOME_KEYS[placement]]);
})));

/**
 * Also need to know the farthest distance between same-finger keys.
 *
 * (Yes, this is a horrible quadratic algorithm).
 */
var maximumDistanceOnSameFinger = KEYS.reduce(function (max, key, i) {
  var finger = FINGERS_PLACEMENTS[i];
  for (var j = 0; j < KEYS.length; j++) {
    if (FINGERS_PLACEMENTS[j] === finger) {
      max = Math.max(max, getDistance(KEYS[i], KEYS[j]));
    }
  }
  return max;
}, 0);

/**
 * And biggest distance between keys on the keyboard (as a pretty good
 * approximation, using the top-left and bottom-left keys).
 */
var maximumDistanceBetweenKeys = getDistance(KEYS[0], KEYS[KEYS.length - 1]);

/**
 * Normalizes `value` within the range defined by `minimum` to `maximum` to a
 * scale from 0 to 1.
 */
function normalize(value, minimum, maximum) {
  // Via: http://stats.stackexchange.com/a/70807
  return (value - minimum) / (maximum - minimum);
}

/**
 * Adjusts the score of a trigram according to any "rolls" that may be present
 * in it.
 *
 * - Inward rolls are favorable, so get strongly boosted.
 * - Outward rolls are unfavorable, so get penalized weakly.
 * - 2-letter rolls that combine alternation score more highly than 2-letter
 *   rolls followed (or preceded) by a key-press on the same hand.
 * - Tighter rolls (ie. ones with less distance) score more favorably.
 */
function getRollMultiplier(trigram, layout) {
  var map = getLayoutLookupMap(layout);
  var letters = trigram.split('');
  var keys = letters.map(function (letter) {
    return KEYS[map[letter].index];
  });

  var handAlternations = 0;
  var inwardRolls = 0;
  var outwardRolls = 0;

  for (var i = 1; i < keys.length; i++) {
    var fingerA = FINGERS_PLACEMENTS[keys[i - 1].id];
    var fingerB = FINGERS_PLACEMENTS[keys[i].id];
    var columnA = keys[i - 1].column;
    var columnB = keys[i].column;
    if (fingerA === fingerB) {
      continue;
    } else if (fingerA <= 4 && fingerB >= 5 || fingerA >= 5 && fingerB <= 4) {
      handAlternations++;
    } else if (columnA < columnB && fingerB <= 4 || columnA > columnB && fingerA >= 5) {
      var distance = getDistance(keys[i - 1], keys[i]);
      inwardRolls += 1 - normalize(distance, 0, maximumDistanceBetweenKeys);
    } else if (columnA > columnB && fingerB <= 4 || columnA < columnB && fingerA >= 5) {
      var distance = getDistance(keys[i - 1], keys[i]);
      outwardRolls += 1 - normalize(distance, 0, maximumDistanceBetweenKeys);
    }
  }
  return 1 + (handAlternations ? handAlternations * -0.2 : 0) + (outwardRolls ? outwardRolls * 0.2 : 0) + (inwardRolls ? inwardRolls * -0.5 : 0);
}

/**
 * Adjusts the score of a trigram according to any contained row jumps.
 *
 * The larger the row jumps, the greater the penalty.
 */
function getRowJumpMultiplier(trigram, layout) {
  var map = getLayoutLookupMap(layout);
  var letters = trigram.split('');
  var keys = letters.map(function (letter) {
    return KEYS[map[letter].index];
  });
  var rowsJumped = 0;
  for (var i = 1; i < keys.length; i++) {
    rowsJumped += Math.abs(keys[i].row - keys[i - 1].row);
  }
  var maximumJump = 10; // Worst case is is row 0 -> row 5 -> row 0.
  var dampener = 0.5; // We don't want this to overwhelm roll boost.
  return 1 + normalize(rowsJumped, 0, maximumJump) * dampener;
}

/**
 * Adjusts the score of a trigram according involving the use of the same finger
 * to make multiple key-presses.
 *
 * The greater the distance between successive key-presses, the greater the
 * penalty.
 */
function getSameFingerMultiplier(trigram, layout) {
  var map = getLayoutLookupMap(layout);
  var letters = trigram.split('');
  var keys = letters.map(function (letter) {
    return map[letter].index;
  });
  return keys.reduce(function (multiplier, key, i) {
    var fingerA = FINGERS_PLACEMENTS[key];
    var fingerB = FINGERS_PLACEMENTS[keys[i - 1]];
    if (fingerA === fingerB) {
      var distance = normalize(getDistance(KEYS[key], KEYS[keys[i - 1]]), 0, maximumDistanceOnSameFinger);
      return multiplier * (distance + 1);
    } else {
      // Different fingers; not our concern here.
      return multiplier;
    }
  }, 1);
}

/**
 * Adjusts the score of a trigram according to the position of the keys
 * involved.
 *
 * In general, keys receive a greater penalty the farther they are from the
 * corresponding finger's position on the home row.
 *
 * The adjustment applied here is fairly weak because there are other, more
 * important factors that need to be weighed more heavily (for example, a
 * distant key becomes more accessible if a previous key in the trigram has
 * caused the hand to float towards the subsequent key's region).
 */
function getPositionMultiplier(trigram, layout) {
  var map = getLayoutLookupMap(layout);
  var letters = trigram.split('');
  var keys = letters.map(function (letter) {
    return map[letter].index;
  });
  var fingers = keys.map(function (key) {
    return FINGERS_PLACEMENTS[key];
  });
  return fingers.reduce(function (multiplier, finger, i) {
    var distance = getDistance(KEYS[keys[i]], KEYS[FINGER_HOME_KEYS[finger]]);
    return multiplier * (normalize(distance, 0, maximumDistanceFromHomeKey) + 1);
  }, 1);
}

/**
 * Modifies the score of a trigram according to the strength of the fingers
 * involved.
 */
function getFingerMultiplier(trigram, layout) {
  var map = getLayoutLookupMap(layout);
  var letters = trigram.split('');
  var fingers = letters.map(function (letter) {
    return FINGERS_PLACEMENTS[map[letter].index];
  });
  return fingers.reduce(function (multiplier, finger) {
    return multiplier * FINGER_STRENGTHS[finger];
  }, 1);
}

var SCORE_MULTIPLIERS = [getRollMultiplier, getRowJumpMultiplier, getSameFingerMultiplier, getPositionMultiplier, getFingerMultiplier];

/**
 * Calculates an effort score for the provided `trigram`. Lower scores are
 * better.
 *
 * Each trigram starts off with an initial score of 1, which is adjusted upwards
 * or downwards from there by a series of "multipliers", each considering a
 * particular factor such as finger strength or row jumps etc.
 */
function scoreTrigram(trigram, layout) {
  return SCORE_MULTIPLIERS.reduce(function (score, scorer) {
    return score * scorer(trigram, layout);
  }, 1);
}

function getSortedFingerCounts(fingerCounts) {
  return Object.keys(fingerCounts).sort(function (a, b) {
    return fingerCounts[b] - fingerCounts[a];
  }).map(function (finger) {
    return [finger, fingerCounts[finger]];
  });
}

function repeat(string, length) {
  while (string.length < length) {
    string += string;
  }
  return string.slice(0, length);
}

function leftAlign(string, width) {
  if (string.length < width) {
    return string + repeat(' ', width - string.length);
  } else {
    return string.slice(0, width);
  }
}

function rightAlign(string, width) {
  if (string.length < width) {
    return repeat(' ', width - string.length) + string;
  } else {
    return string.slice(0, width);
  }
}

function getBar(count, total, width) {
  return '|' + repeat('-', Math.floor(count / total * width)) + 'o' + repeat(' ', Math.floor((total - count) / total * width)) + '|';
}

function printHistogram(rows, total) {
  var terminalWidth = process.stdout.columns || 80;
  var labelWidth = Math.max.apply(Math, _toConsumableArray(rows.map(function (row) {
    return row.label.length;
  }))) + 2;
  var counts = rows.map(function (row) {
    return formatNumber(row.count);
  });
  var countWidth = Math.max.apply(Math, _toConsumableArray(counts.map(function (count) {
    return count.length;
  }))) + 1;
  var percentages = rows.map(function (row) {
    return getPercentage(row.count, total);
  });
  var percentageWidth = Math.max.apply(Math, _toConsumableArray(percentages.map(function (percentage) {
    return percentage.length;
  }))) + 3;
  var barWidth = Math.max(10, Math.min(80, terminalWidth - labelWidth - countWidth - percentageWidth - 2));

  rows.forEach(function (row, i) {
    print(rightAlign(row.label + ': ', labelWidth));
    print(rightAlign(counts[i] + ' ', countWidth));
    print(leftAlign('(' + percentages[i] + ') ', percentageWidth));
    log(getBar(row.count, total, barWidth));
  });
}

function printLayoutStats(layout, corpus) {
  printHeading(layout.name + ' layout stats:');
  var totalCount = 0;
  var fingerCounts = {};
  var rowCounts = {};
  var map = getLayoutLookupMap(layout);
  var unigrams = corpus.split('').filter(function (letter) {
    return N_GRAM_REGEXP.test(letter);
  });
  unigrams.forEach(function (letter) {
    var keyIndex = map[getLetterForDisplay(letter)];
    var fingerIndex = FINGERS_PLACEMENTS[keyIndex.index]; // Ignore Shift for now.
    fingerCounts[fingerIndex] = fingerCounts[fingerIndex] || 0;
    fingerCounts[fingerIndex]++;
    var row = KEYS[keyIndex.index].row;
    rowCounts[row] = rowCounts[row] || 0;
    rowCounts[row]++;
    totalCount++;
  });

  printHeading('Finger utilization:');
  printHistogram(Object.keys(fingerCounts).map(function (finger) {
    return {
      label: FINGER_NAMES[finger],
      count: fingerCounts[finger]
    };
  }), totalCount);

  printHeading('Hand utilization:');
  var hands = Object.keys(fingerCounts).reduce(function (hands, finger) {
    var count = fingerCounts[finger];
    var hand = undefined;
    // Exclude the thumb.
    if (finger < 4) {
      hand = 'left';
    } else if (finger > 5) {
      hand = 'right';
    } else {
      return hands;
    }
    hands[hand] = hands[hand] || 0;
    hands[hand] += count;
    return hands;
  }, {});
  printHistogram([{ label: 'Left', count: hands.left }, { label: 'Right', count: hands.right }], totalCount);

  printHeading('Row usage:');
  printHistogram(Object.keys(rowCounts).map(function (row) {
    return {
      label: 'Row ' + row + ' (' + ROWS[row] + ')',
      count: rowCounts[row]
    };
  }), totalCount);

  printHeading('Effort (per trigram):');

  var _getNGramFrequencies = getNGramFrequencies(corpus, 3);

  var trigrams = _getNGramFrequencies.nGrams;

  var sortedTrigrams = getSortedNGrams(trigrams);
  var totalEffort = 0;
  sortedTrigrams.slice(0, 50).forEach(function (_ref) {
    var _ref2 = _slicedToArray(_ref, 2);

    var key = _ref2[0];
    var count = _ref2[1];

    var percentage = getPercentage(count, totalCount);
    var score = scoreTrigram(key, layout);
    var total = score * count;
    totalEffort += total;
    log(getLettersForDisplay(key) + ': ' + formatNumber(count) + ' (' + percentage + ') [score: ' + formatNumber(score, 4) + ', total: ' + formatNumber(total) + ']');
  });
  log('Total effort: ' + formatNumber(totalEffort));

  printHeading('Summary:');
  SCORE_MULTIPLIERS.forEach(function (multiplier) {
    var overallCount = 0;
    var total = sortedTrigrams.slice(0, 50).reduce(function (total, _ref3) {
      var _ref32 = _slicedToArray(_ref3, 2);

      var key = _ref32[0];
      var count = _ref32[1];

      var score = multiplier(key, layout);
      overallCount += count;
      return total + score * count;
    }, 0);
    log(multiplier.name + ': ' + formatNumber(total / overallCount, 4));
  });
}

var HEADING = '----------------------------------------' + '----------------------------------------';

function printHeading(string) {
  var length = string.length;
  log('\n' + string);
  log(HEADING.slice(0, length));
}

function getLettersForDisplay(letters) {
  return letters.split('').map(function (letter) {
    return getLetterForDisplay(letter);
  }).join('');
}

function getLetterForDisplay(letter) {
  if (letter === ' ') {
    return '␣';
  } else if (letter === '\n') {
    return '↩';
  } else if (letter === '\t') {
    return '⇥';
  } else {
    return letter;
  }
}

function escapeRegExp(string) {
  // From:
  // https://developer.mozilla.org/en/docs/Web/JavaScript/Guide/Regular_Expressions
  return string.replace(/[.*+?^${}()|[\]\\]/g, '\\$&');
}

function getPrintableASCII() {
  var ascii = '';
  for (var i = 32; i <= 127; i++) {
    ascii += String.fromCharCode(i);
  }
  return ascii;
}

var N_GRAM_CHARACTER_CLASS = '[' + escapeRegExp(getPrintableASCII()) + ']';
var N_GRAM_REGEXP = new RegExp('^' + N_GRAM_CHARACTER_CLASS + '+$', 'gi');
var WHITESPACE_REGEXP = new RegExp('[ \t\n]');

function isValidNGram(nGram) {
  return !WHITESPACE_REGEXP.test(nGram) && N_GRAM_REGEXP.test(nGram);
}

function getNGramFrequencies(corpus, n) {
  var nGrams = {};
  var totalCount = 0;
  for (var i = 0, max = corpus.length - n + 1; i < max; i++) {
    var _nGram = corpus.substr(i, n);
    if (isValidNGram(_nGram)) {
      nGrams[_nGram] = nGrams[_nGram] || 0;
      nGrams[_nGram]++;
      totalCount++;
    }
  }
  return { nGrams: nGrams, totalCount: totalCount };
}

function getSortedNGrams(nGrams) {
  return Object.keys(nGrams).sort(function (a, b) {
    var diff = nGrams[b] - nGrams[a];
    if (diff) {
      return diff;
    } else {
      // Count is same; use lexicographical sort to break tie.
      return a < b ? -1 : 1;
    }
  }).map(function (key) {
    return [key, nGrams[key]];
  });
}

function formatNumber(number) {
  var precision = arguments.length <= 1 || arguments[1] === undefined ? 0 : arguments[1];

  var _number$toFixed$split = number.toFixed(precision).split('.');

  // Borrow trick from ActiveSupport, using positive lookahead (?=) and negative
  // lookahead (?!).

  var _number$toFixed$split2 = _slicedToArray(_number$toFixed$split, 2);

  var integer = _number$toFixed$split2[0];
  var decimal = _number$toFixed$split2[1];
  var delimited = integer.replace(/(\d)(?=(\d\d\d)+(?!\d))/g, function (_, group) {
    return group + ',';
  });

  if (decimal) {
    return [delimited, decimal].join('.');
  } else {
    return delimited;
  }
}

function getPercentage(dividend, divisor) {
  var precision = arguments.length <= 2 || arguments[2] === undefined ? 2 : arguments[2];

  return (dividend / divisor * 100).toFixed(precision) + '%';
}

/**
 * Returns the "fitness" of a layout for typing the corpus represented by
 * `trigrams`.
 *
 * "fitness" corresponds to total typing effort, so lower scores are better.
 */
function getFitness(layout, trigrams) {
  var totalEffort = 0;
  trigrams.slice(0, 100).forEach(function (_ref4) {
    var _ref42 = _slicedToArray(_ref4, 2);

    var key = _ref42[0];
    var count = _ref42[1];

    var score = scoreTrigram(key, layout);
    var total = score * count;
    totalEffort += total;
  });
  return totalEffort;
}

/**
 * Returns the count of keys (pairs of keys) that should be swapped, with some
 * random probability.
 */
function getSwapCount() {
  var random = Math.random();
  if (random < 0.9) {
    return 1;
  } else if (random < 0.99) {
    return 2;
  } else {
    return 3;
  }
}

/**
 * Given a `layout`, generate a digest that serves as (an almost certainly)
 * unique fingerprint for it, and can be used to detect duplicate layouts.
 */
function getLayoutDigest(layout) {
  return layout.keys.reduce(function (digest, key) {
    return digest + key;
  }, '');
}

/**
 * Pretty-prints `layout` in human-readable form.
 */
function printLayout(layout) {
  var lastRow = null;
  KEYS.forEach(function (_ref5, i) {
    var row = _ref5.row;
    var column = _ref5.column;

    var entry = layout.keys[i];
    var key = Array.isArray(entry) ? entry[0] : entry;
    if (row !== lastRow) {
      lastRow = row;
      if (row) {
        print('\n');
      }
    }
    if (column) {
      print('  ');
    }
    if (/^F\d+$/.test(key)) {
      // This makes row 0 too long, but so be it.
      print(key.slice(0, 3));
    } else {
      print(key.slice(0, 1));
    }
  });
  print('\n');
}

function swapKeys(layout, sourceIndex, targetIndex) {
  var temp = layout.keys[targetIndex];
  layout.keys[targetIndex] = layout.keys[sourceIndex];
  layout.keys[sourceIndex] = temp;
}

function checkNoOps(layout, seen, sourceIndex, targetIndex) {
  return sourceIndex !== targetIndex;
}

function checkDuplicates(layout, seen, sourceIndex, targetIndex) {
  // Must apply the mutation in order to get the new digest.
  var testLayout = {
    keys: layout.keys.slice(),
    name: 'Test layout ' + layoutCounter++
  };
  swapKeys(testLayout, sourceIndex, targetIndex);
  return !seen[getLayoutDigest(testLayout)];
}

function checkMaskedKeys(layout, seen, sourceIndex, targetIndex) {
  return MASK[sourceIndex] !== 1 && MASK[targetIndex] !== 1;
  return true;
}

function anneal(delta, iteration, iterationCount) {
  // Based on explanation at: http://mkweb.bcgsc.ca/carpalx/?simulated_annealing
  var t = 250000 * Math.exp(-iteration * 10 / iterationCount);
  var p = Math.exp(-delta / t);
  return Math.random() < p;
}

function now() {
  var _process$hrtime = process.hrtime();

  var _process$hrtime2 = _slicedToArray(_process$hrtime, 2);

  var seconds = _process$hrtime2[0];
  var nanoseconds = _process$hrtime2[1];

  return seconds + nanoseconds / 1000000000;
}

function optimize(layout, sortedTrigrams, iterationCount) {
  log('Optimizing ' + layout.name + ':');
  var start = now();
  var fitness = getFitness(layout, sortedTrigrams);
  var bestFitness = fitness;
  var bestLayout = layout;
  log('Starting fitness: ' + formatNumber(fitness, 2));
  var seen = _defineProperty({}, getLayoutDigest(layout), true);
  for (var i = 0; i < iterationCount; i++) {
    log('Iteration ' + formatNumber(i) + ':');
    var evolvedLayout = evolve(layout, seen);
    if (!evolvedLayout) {
      continue;
    }
    printLayout(evolvedLayout);
    var evolvedFitness = getFitness(evolvedLayout, sortedTrigrams);
    print('Fitness: ' + formatNumber(evolvedFitness, 2) + ' ');
    if (evolvedFitness < fitness) {
      log('[better » accepting]');
      fitness = evolvedFitness;
      layout = evolvedLayout;
    } else if (anneal(evolvedFitness - fitness, i, iterationCount)) {
      log('[worse » accepting]');
      fitness = evolvedFitness;
      layout = evolvedLayout;
    } else {
      log('[worse » rejecting]');
    }

    if (fitness < bestFitness) {
      bestFitness = fitness;
      bestLayout = layout;
    }
  }
  log('Final layout:');
  printLayout(bestLayout);
  log('Final fitness: ' + formatNumber(bestFitness, 2));
  var finish = now();
  var elapsed = finish - start;
  console.log('Elapsed time: ' + formatNumber(elapsed, 2) + 's');
}

/**
 * Takes `layout` and applies a random mutation to it, returning a new layout.
 *
 * Takes a `seen` hash recording which layouts have previously been considered.
 * Updates the hash.
 */
function evolve(layout, seen) {
  var evolved = {
    // NOTE: not a deep clone, so for now we swap shifted/unshifted values of
    // each key together in lock-step.
    keys: layout.keys.slice(),
    name: 'Random layout ' + layoutCounter++
  };

  // Will swap 1, 2 or 3 pairs of keys.
  var swapCount = getSwapCount();
  var attemptCount = 0;

  var _loop = function () {
    attemptCount++;
    var sourceIndex = Math.floor(Math.random() * evolved.keys.length);
    var targetIndex = Math.floor(Math.random() * evolved.keys.length);
    var valid = [checkNoOps, checkDuplicates, checkMaskedKeys].every(function (validator) {
      return validator(evolved, seen, sourceIndex, targetIndex);
    });
    if (valid) {
      swapKeys(evolved, sourceIndex, targetIndex);
      swapCount--;
    } else if (attemptCount > 1000) {
      log('Likely deadlock in evolve() (bailing)');
      return {
        v: null
      };
    }
  };

  while (swapCount) {
    var _ret2 = _loop();

    if (typeof _ret2 === 'object') return _ret2.v;
  }

  seen[getLayoutDigest(evolved)] = true;
  return evolved;
}

function printCorpusStats(corpus) {
  log('Corpus length: ' + formatNumber(corpus.length) + ' bytes');

  var _getNGramFrequencies2 = getNGramFrequencies(corpus, 1);

  var unigrams = _getNGramFrequencies2.nGrams;
  var unigramCount = _getNGramFrequencies2.totalCount;

  var _getNGramFrequencies3 = getNGramFrequencies(corpus, 2);

  var bigrams = _getNGramFrequencies3.nGrams;
  var bigramCount = _getNGramFrequencies3.totalCount;

  var _getNGramFrequencies4 = getNGramFrequencies(corpus, 3);

  var trigrams = _getNGramFrequencies4.nGrams;
  var trigramCount = _getNGramFrequencies4.totalCount;

  var sortedUnigrams = getSortedNGrams(unigrams).filter(function (_ref6) {
    var _ref62 = _slicedToArray(_ref6, 2);

    var key = _ref62[0];
    var count = _ref62[1];
    return key !== ' ';
  });
  var sortedBigrams = getSortedNGrams(bigrams);
  var sortedTrigrams = getSortedNGrams(trigrams);

  [{ label: 'Unigrams', nGrams: sortedUnigrams, top: 100, totalCount: unigramCount }, { label: 'Bigrams', nGrams: sortedBigrams, top: 50, totalCount: bigramCount }, { label: 'Trigrams', nGrams: sortedTrigrams, top: 50, totalCount: trigramCount }].forEach(function (_ref7) {
    var label = _ref7.label;
    var nGrams = _ref7.nGrams;
    var top = _ref7.top;
    var totalCount = _ref7.totalCount;

    var total = nGrams.length;
    top = Math.min(top, total);
    printHeading(label + ' by frequency (top ' + top + ' of ' + formatNumber(total) + '):');
    nGrams.slice(0, top).forEach(function (_ref8) {
      var _ref82 = _slicedToArray(_ref8, 2);

      var key = _ref82[0];
      var count = _ref82[1];

      var percentage = getPercentage(count, totalCount);
      log(getLettersForDisplay(key) + ': ' + formatNumber(count) + ' (' + percentage + ')');
    });
  });

  var overview = sortedUnigrams.map(function (_ref9) {
    var _ref92 = _slicedToArray(_ref9, 2);

    var key = _ref92[0];
    var count = _ref92[1];
    return getLetterForDisplay(key);
  }).filter(function (letter) {
    return letter.length === 1;
  }).join('');
  printHeading('Unigrams frequency overview:');
  log(overview);
}

function getRandomLayout() {
  var layout = LAYOUTS.QWERTY;
  console.log('Creating random layout...');
  var seen = _defineProperty({}, getLayoutDigest(layout), true);
  for (var i = 0; i < 1000; i++) {
    layout = evolve(layout, seen);
  }
  return layout;
}

(function callee$0$0() {
  var json, common, argv, getCorpus, command, corpus, layout, _getNGramFrequencies5, trigrams, sortedTrigrams;

  return regeneratorRuntime.async(function callee$0$0$(context$1$0) {
    while (1) switch (context$1$0.prev = context$1$0.next) {
      case 0:
        getCorpus = function getCorpus() {
          var corpusPath, corpus;
          return regeneratorRuntime.async(function getCorpus$(context$2$0) {
            while (1) switch (context$2$0.prev = context$2$0.next) {
              case 0:
                corpusPath = _path2['default'].join('yak', 'corpus.txt');
                context$2$0.next = 3;
                return regeneratorRuntime.awrap(readFile(corpusPath));

              case 3:
                corpus = context$2$0.sent;
                return context$2$0.abrupt('return', corpus.toString().toLowerCase());

              case 5:
              case 'end':
                return context$2$0.stop();
            }
          }, null, this);
        };

        common = function common(yargsish) {
          return yargsish.help('h').alias('h', 'help').version(json.version).epilog(json.homepage).strict().argv;
        };

        json = require('../package');
        argv = common(_yargs2['default'].usage('Usage: $0 <command> [options...]').command('corpus-stats', 'show corpus stats', function (yargs) {
          argv = common(yargs.reset().usage('Usage: $0 corpus-stats'));
        }).command('help', 'Show help').command('layout-stats', 'show layout stats', function (yargs) {
          argv = common(yargs.reset().usage('Usage: $0 layout-stats [layout]'));
        }).command('optimize', 'produce optimized keyboard layout', function (yargs) {
          argv = common(yargs.reset().usage('Usage: $0 optimize [layout]')['default']('iteration-count', 10000).alias('c', 'iteration-count'));
        }).demand(1, 'must provide a valid command'));
        command = argv._[0];

        if (!(command === 'corpus-stats')) {
          context$1$0.next = 12;
          break;
        }

        context$1$0.next = 8;
        return regeneratorRuntime.awrap(getCorpus());

      case 8:
        context$1$0.t0 = context$1$0.sent;
        printCorpusStats(context$1$0.t0);
        context$1$0.next = 42;
        break;

      case 12:
        if (!(command === 'layout-stats')) {
          context$1$0.next = 22;
          break;
        }

        context$1$0.next = 15;
        return regeneratorRuntime.awrap(getCorpus());

      case 15:
        corpus = context$1$0.sent;
        layout = (argv._[1] || 'QWERTY').toUpperCase();

        if (layout in LAYOUTS) {
          context$1$0.next = 19;
          break;
        }

        throw new Error('Unknown layout: ' + layout);

      case 19:
        printLayoutStats(LAYOUTS[layout], corpus);
        context$1$0.next = 42;
        break;

      case 22:
        if (!(command === 'optimize')) {
          context$1$0.next = 41;
          break;
        }

        context$1$0.next = 25;
        return regeneratorRuntime.awrap(getCorpus());

      case 25:
        corpus = context$1$0.sent;
        _getNGramFrequencies5 = getNGramFrequencies(corpus, 3);
        trigrams = _getNGramFrequencies5.nGrams;
        sortedTrigrams = getSortedNGrams(trigrams);
        layout = undefined;

        if (!argv._[1]) {
          context$1$0.next = 37;
          break;
        }

        layout = argv._[1].toUpperCase();

        if (layout in LAYOUTS) {
          context$1$0.next = 34;
          break;
        }

        throw new Error('Unknown layout: ' + layout);

      case 34:
        layout = LAYOUTS[layout];
        context$1$0.next = 38;
        break;

      case 37:
        layout = getRandomLayout();

      case 38:
        optimize(layout, sortedTrigrams, argv.iterationCount);
        context$1$0.next = 42;
        break;

      case 41:
        _yargs2['default'].showHelp();

      case 42:
      case 'end':
        return context$1$0.stop();
    }
  }, null, this);
})();